#!/usr/bin/python
import hashlib
import random

# find a collision over the first n hexidecimal characters of SHA-1
n=12

# distinguished points have d 0's
d=(n/2)-3
# dcomp is string of d 0's
dcomp=("0000000000000000000000000000")[:d]

# we terminate a chain when it reaches length maxk = 20 times expected trail length = 20 * (4^d)
maxk=20*(1<<(4*d))

# we start with an empty table of endpoints
#  dict() is an associative map
#  maps: endpoint  to  triple(endpoint,startpoint,traillen)
eptbl=dict()

# randomize
random.seed()

print n
print d
print dcomp

# generate trails
hl=hashlib
while (1):
	r=random.getrandbits(32)
	# startpoint sp is first n characters of SHA-1 hash of the ascii string representing a 32-bit integer r
	sp=hl.sha1(str(r)).hexdigest()[:n]
	k=0
	p=sp
	# start with p = sp
	# iterate p=SHA-1(p)[:n] until p is a distinguished point (i.e., p[:d]==dcomp) or trail length reaches maxk
	while not (p[:d]==dcomp):
		k=k+1
		p=hl.sha1(p).hexdigest()[:n]
		if (k > maxk):
			break
	# check if length maxk was reached
	if not (p[:d]==dcomp):
		print [p,sp,k,'too long']
		continue
	# new trails found
	triple=[p,sp,k]
	print triple
	# check if endpoint was already in table
	if (p in eptbl):
		triple2=eptbl[p]
		break
	# if not then insert triple into table
	eptbl[p]=triple

# found two trails with same endpoint
print ""
print triple
print triple2
	
p1=triple[1]
p2=triple2[1]
k1=triple[2]
k2=triple2[2]
s1=p1
s2=p2

# reduce the longest trail to the same size as the shortest
while (k1 > k2):
	k1=k1-1
	s1=p1
	p1=hl.sha1(p1).hexdigest()[:n]

while (k2 > k1):
	k2=k2-1
	s2=p2
	p2=hl.sha1(p2).hexdigest()[:n]

# walk both trails in-step to find actual collision point
while not (p1==p2):
	k1=k1-1
	s1=p1
	p1=hl.sha1(p1).hexdigest()[:n]
	k2=k2-1
	s2=p2
	p2=hl.sha1(p2).hexdigest()[:n]
	if (k1==0):
		print "try again"
		break

print [s1,s2]
print p1
	